# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Image model specification."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools

from tensorflow_examples.lite.model_maker.core import compat
from tensorflow_examples.lite.model_maker.core.api import mm_export
from tensorflow_examples.lite.model_maker.core.task import configs
from tensorflow_examples.lite.model_maker.core.task.model_spec import util


@mm_export('image_classifier.ModelSpec')
class ImageModelSpec(object):
  """A specification of image model."""

  mean_rgb = [0.0]
  stddev_rgb = [255.0]

  def __init__(self,
               uri,
               compat_tf_versions=None,
               input_image_shape=None,
               name=''):
    """Initializes a new instance of the `ImageModelSpec` class.

    Args:
      uri: str, URI to the pretrained model.
      compat_tf_versions: list of int, compatible TF versions.
      input_image_shape: list of int, input image shape. Default: [224, 224].
      name: str, model spec name.
    """
    self.uri = uri
    self.compat_tf_versions = compat.get_compat_tf_versions(compat_tf_versions)
    self.name = name

    if input_image_shape is None:
      input_image_shape = [224, 224]
    self.input_image_shape = input_image_shape

  def get_default_quantization_config(self, representative_data):
    """Gets the default quantization configuration."""
    config = configs.QuantizationConfig.for_int8(representative_data)
    return config


mobilenet_v2_spec = functools.partial(
    ImageModelSpec,
    uri='https://tfhub.dev/google/tf2-preview/mobilenet_v2/feature_vector/4',
    compat_tf_versions=2,
    name='mobilenet_v2')
mobilenet_v2_spec.__doc__ = util.wrap_doc(
    ImageModelSpec,
    'Creates MobileNet v2 model spec. See also: `tflite_model_maker.image_classifier.ModelSpec`.'
)
mm_export('image_classifier.MobileNetV2Spec').export_constant(
    __name__, 'mobilenet_v2_spec')

resnet_50_spec = functools.partial(
    ImageModelSpec,
    uri='https://tfhub.dev/google/imagenet/resnet_v2_50/feature_vector/4',
    compat_tf_versions=2,
    name='resnet_50')
resnet_50_spec.__doc__ = util.wrap_doc(
    ImageModelSpec,
    'Creates ResNet 50 model spec. See also: `tflite_model_maker.image_classifier.ModelSpec`.'
)
mm_export('image_classifier.Resnet50Spec').export_constant(
    __name__, 'resnet_50_spec')

efficientnet_lite0_spec = functools.partial(
    ImageModelSpec,
    uri='https://tfhub.dev/tensorflow/efficientnet/lite0/feature-vector/2',
    compat_tf_versions=[1, 2],
    name='efficientnet_lite0')
efficientnet_lite0_spec.__doc__ = util.wrap_doc(
    ImageModelSpec,
    'Creates EfficientNet-Lite0 model spec. See also: `tflite_model_maker.image_classifier.ModelSpec`.'
)
mm_export('image_classifier.EfficientNetLite0Spec').export_constant(
    __name__, 'efficientnet_lite0_spec')

efficientnet_lite1_spec = functools.partial(
    ImageModelSpec,
    uri='https://tfhub.dev/tensorflow/efficientnet/lite1/feature-vector/2',
    compat_tf_versions=[1, 2],
    input_image_shape=[240, 240],
    name='efficientnet_lite1')
efficientnet_lite1_spec.__doc__ = util.wrap_doc(
    ImageModelSpec,
    'Creates EfficientNet-Lite1 model spec. See also: `tflite_model_maker.image_classifier.ModelSpec`.'
)
mm_export('image_classifier.EfficientNetLite1Spec').export_constant(
    __name__, 'efficientnet_lite1_spec')

efficientnet_lite2_spec = functools.partial(
    ImageModelSpec,
    uri='https://tfhub.dev/tensorflow/efficientnet/lite2/feature-vector/2',
    compat_tf_versions=[1, 2],
    input_image_shape=[260, 260],
    name='efficientnet_lite2')
efficientnet_lite2_spec.__doc__ = util.wrap_doc(
    ImageModelSpec,
    'Creates EfficientNet-Lite2 model spec. See also: `tflite_model_maker.image_classifier.ModelSpec`.'
)
mm_export('image_classifier.EfficientNetLite2Spec').export_constant(
    __name__, 'efficientnet_lite2_spec')

efficientnet_lite3_spec = functools.partial(
    ImageModelSpec,
    uri='https://tfhub.dev/tensorflow/efficientnet/lite3/feature-vector/2',
    compat_tf_versions=[1, 2],
    input_image_shape=[280, 280],
    name='efficientnet_lite3')
efficientnet_lite3_spec.__doc__ = util.wrap_doc(
    ImageModelSpec,
    'Creates EfficientNet-Lite3 model spec. See also: `tflite_model_maker.image_classifier.ModelSpec`.'
)
mm_export('image_classifier.EfficientNetLite3Spec').export_constant(
    __name__, 'efficientnet_lite3_spec')

efficientnet_lite4_spec = functools.partial(
    ImageModelSpec,
    uri='https://tfhub.dev/tensorflow/efficientnet/lite4/feature-vector/2',
    compat_tf_versions=[1, 2],
    input_image_shape=[300, 300],
    name='efficientnet_lite4')
efficientnet_lite4_spec.__doc__ = util.wrap_doc(
    ImageModelSpec,
    'Creates EfficientNet-Lite4 model spec. See also: `tflite_model_maker.image_classifier.ModelSpec`.'
)
mm_export('image_classifier.EfficientNetLite4Spec').export_constant(
    __name__, 'efficientnet_lite4_spec')



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Grid Masking Augmentation Reference: https://arxiv.org/abs/2001.04086."""
import math

import tensorflow as tf
from tensorflow_addons import image as image_ops


class GridMask(object):
  """GridMask class for grid masking augmentation."""

  def __init__(self,
               prob=0.6,
               ratio=0.6,
               rotate=10,
               gridmask_size_ratio=0.5,
               fill=1,
               interpolation="BILINEAR"):
    """initialization.

    Args:
      prob: probablity of occurance.
      ratio: grid mask ratio i.e if 0.5 grid and spacing will be equal.
      rotate: Rotation of grid mesh.
      gridmask_size_ratio: Grid mask size, grid to image size ratio.
      fill: Fill value for grids.
      interpolation: Interpolation method for rotation.
    """
    self.prob = prob
    self.ratio = ratio
    self.rotate = rotate
    self.gridmask_size_ratio = gridmask_size_ratio
    self.fill = fill
    self.interpolation = interpolation

  @tf.function
  def random_rotate(self, mask):
    """Randomly rotates mask on given range."""

    angle = self.rotate * tf.random.normal([], -1, 1)
    angle = math.pi * angle / 180
    return image_ops.rotate(mask, angle, interpolation=self.interpolation)

  @staticmethod
  def crop(mask, h, w):
    """crops in middle of mask and image corners."""
    ww = hh = tf.shape(mask)[0]
    mask = mask[(hh - h) // 2:(hh - h) // 2 + h,
                (ww - w) // 2:(ww - w) // 2 + w,]
    return mask

  @tf.function
  def mask(self, h, w):
    """mask helper function for initializing grid mask of required size."""
    h = tf.cast(h, tf.float32)
    w = tf.cast(w, tf.float32)
    mask_w = mask_h = tf.cast(
        tf.cast(
            (self.gridmask_size_ratio + 1), tf.float32) * tf.math.maximum(h, w),
        tf.int32)
    self.mask_w = mask_w
    mask = tf.zeros(shape=[mask_h, mask_w], dtype=tf.int32)
    gridblock = tf.random.uniform(
        shape=[],
        minval=int(tf.math.minimum(h * 0.5, w * 0.3)),
        maxval=int(tf.math.maximum(h * 0.5, w * 0.3)) + 1,
        dtype=tf.int32)

    if self.ratio == 1:
      length = tf.random.uniform(
          shape=[], minval=1, maxval=gridblock + 1, dtype=tf.int32)
    else:
      length = tf.cast(
          tf.math.minimum(
              tf.math.maximum(
                  int(tf.cast(gridblock, tf.float32) * self.ratio + 0.5), 1),
              gridblock - 1), tf.int32)

    for _ in range(2):
      start_w = tf.random.uniform(
          shape=[], minval=0, maxval=gridblock + 1, dtype=tf.int32)
      for i in range(mask_w // gridblock):
        start = gridblock * i + start_w
        end = tf.math.minimum(start + length, mask_w)
        indices = tf.reshape(tf.range(start, end), [end - start, 1])
        updates = (
            tf.ones(shape=[end - start, mask_w], dtype=tf.int32) * self.fill)
        mask = tf.tensor_scatter_nd_update(mask, indices, updates)
      mask = tf.transpose(mask)

    return mask

  def __call__(self, image, label):
    """Masks input image tensor with random grid mask."""
    h = tf.shape(image)[0]
    w = tf.shape(image)[1]
    grid = self.mask(h, w)
    grid = self.random_rotate(grid)
    mask = self.crop(grid, h, w)
    mask = tf.cast(mask, image.dtype)
    mask = tf.reshape(mask, (h, w))
    mask = (tf.expand_dims(mask, -1) if image._rank() != mask._rank() else mask)
    occur = tf.random.normal([], 0, 1) < self.prob
    image = tf.cond(occur, lambda: image * mask, lambda: image)
    return image, label


def gridmask(image,
             boxes,
             prob=0.5,
             ratio=0.6,
             rotate=10,
             gridmask_size_ratio=0.5,
             fill=1):
  """Callable instance of GridMask and transforms input image."""
  gridmask_obj = GridMask(
      prob=prob,
      ratio=ratio,
      rotate=rotate,
      gridmask_size_ratio=gridmask_size_ratio,
      fill=fill)
  image, boxes = gridmask_obj(image, boxes)
  return image, boxes



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Model specification."""

import functools
import inspect

from tensorflow_examples.lite.model_maker.core.api import mm_export
from tensorflow_examples.lite.model_maker.core.task.model_spec import audio_spec
from tensorflow_examples.lite.model_maker.core.task.model_spec import image_spec
from tensorflow_examples.lite.model_maker.core.task.model_spec import object_detector_spec
from tensorflow_examples.lite.model_maker.core.task.model_spec import recommendation_spec
from tensorflow_examples.lite.model_maker.core.task.model_spec import text_spec


# A dict for model specs to make it accessible by string key.
MODEL_SPECS = {
    # Image classification
    'efficientnet_lite0': image_spec.efficientnet_lite0_spec,
    'efficientnet_lite1': image_spec.efficientnet_lite1_spec,
    'efficientnet_lite2': image_spec.efficientnet_lite2_spec,
    'efficientnet_lite3': image_spec.efficientnet_lite3_spec,
    'efficientnet_lite4': image_spec.efficientnet_lite4_spec,
    'mobilenet_v2': image_spec.mobilenet_v2_spec,
    'resnet_50': image_spec.resnet_50_spec,

    # Text classification
    'average_word_vec': text_spec.AverageWordVecModelSpec,
    'bert': text_spec.BertModelSpec,
    'bert_classifier': text_spec.BertClassifierModelSpec,
    'mobilebert_classifier': text_spec.mobilebert_classifier_spec,

    # Question answering
    'bert_qa': text_spec.BertQAModelSpec,
    'mobilebert_qa': text_spec.mobilebert_qa_spec,
    'mobilebert_qa_squad': text_spec.mobilebert_qa_squad_spec,

    # Audio classification
    'audio_browser_fft': audio_spec.BrowserFFTSpec,
    'audio_teachable_machine': audio_spec.BrowserFFTSpec,
    'audio_yamnet': audio_spec.YAMNetSpec,

    # Recommendation
    'recommendation': recommendation_spec.RecommendationSpec,

    # Object detection
    'efficientdet_lite0': object_detector_spec.efficientdet_lite0_spec,
    'efficientdet_lite1': object_detector_spec.efficientdet_lite1_spec,
    'efficientdet_lite2': object_detector_spec.efficientdet_lite2_spec,
    'efficientdet_lite3': object_detector_spec.efficientdet_lite3_spec,
    'efficientdet_lite4': object_detector_spec.efficientdet_lite4_spec,
}

# List constants for supported models.
IMAGE_CLASSIFICATION_MODELS = [
    'efficientnet_lite0', 'efficientnet_lite1', 'efficientnet_lite2',
    'efficientnet_lite3', 'efficientnet_lite4', 'mobilenet_v2', 'resnet_50'
]
TEXT_CLASSIFICATION_MODELS = [
    'bert_classifier', 'average_word_vec', 'mobilebert_classifier'
]
QUESTION_ANSWER_MODELS = ['bert_qa', 'mobilebert_qa', 'mobilebert_qa_squad']
AUDIO_CLASSIFICATION_MODELS = [
    'audio_browser_fft', 'audio_teachable_machine', 'audio_yamnet'
]
RECOMMENDATION_MODELS = [
    'recommendation',
]
OBJECT_DETECTION_MODELS = [
    'efficientdet_lite0',
    'efficientdet_lite1',
    'efficientdet_lite2',
    'efficientdet_lite3',
    'efficientdet_lite4',
]
mm_export('model_spec.IMAGE_CLASSIFICATION_MODELS').export_constant(
    __name__, 'IMAGE_CLASSIFICATION_MODELS')
mm_export('model_spec.TEXT_CLASSIFICATION_MODELS').export_constant(
    __name__, 'TEXT_CLASSIFICATION_MODELS')
mm_export('model_spec.QUESTION_ANSWER_MODELS').export_constant(
    __name__, 'QUESTION_ANSWER_MODELS')
mm_export('model_spec.AUDIO_CLASSIFICATION_MODELS').export_constant(
    __name__, 'AUDIO_CLASSIFICATION_MODELS')
mm_export('model_spec.RECOMMENDATION_MODELS').export_constant(
    __name__, 'RECOMMENDATION_MODELS')
mm_export('model_spec.OBJECT_DETECTION_MODELS').export_constant(
    __name__, 'OBJECT_DETECTION_MODELS')


@mm_export('model_spec.get')
def get(spec_or_str, *args, **kwargs):
  """Gets model spec by name or instance, and init with args and kwarges."""
  if isinstance(spec_or_str, str):
    model_spec = MODEL_SPECS[spec_or_str]
  else:
    model_spec = spec_or_str

  if inspect.isclass(model_spec) or inspect.isfunction(
      model_spec) or isinstance(model_spec, functools.partial):
    return model_spec(*args, **kwargs)
  else:
    return model_spec



# Copyright 2020 Google Research. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Base target assigner module.

The job of a TargetAssigner is, for a given set of anchors (bounding boxes) and
groundtruth detections (bounding boxes), to assign classification and regression
targets to each anchor as well as weights to each anchor (specifying, e.g.,
which anchors should not contribute to training loss).

It assigns classification/regression targets by performing the following steps:
1) Computing pairwise similarity between anchors and groundtruth boxes using a
  provided RegionSimilarity Calculator
2) Computing a matching based on the similarity matrix using a provided Matcher
3) Assigning regression targets based on the matching and a provided BoxCoder
4) Assigning classification targets based on the matching and groundtruth labels

Note that TargetAssigners only operate on detections from a single
image at a time, so any logic for applying a TargetAssigner to multiple
images must be handled externally.
"""
import tensorflow.compat.v1 as tf

from tensorflow_examples.lite.model_maker.third_party.efficientdet.object_detection import box_list
from tensorflow_examples.lite.model_maker.third_party.efficientdet.object_detection import shape_utils


KEYPOINTS_FIELD_NAME = 'keypoints'


class TargetAssigner(object):
  """Target assigner to compute classification and regression targets."""

  def __init__(self, similarity_calc, matcher, box_coder,
               negative_class_weight=1.0, unmatched_cls_target=None):
    """Construct Object Detection Target Assigner.

    Args:
      similarity_calc: a RegionSimilarityCalculator
      matcher: Matcher used to match groundtruth to anchors.
      box_coder: BoxCoder used to encode matching groundtruth boxes with
        respect to anchors.
      negative_class_weight: classification weight to be associated to negative
        anchors (default: 1.0). The weight must be in [0., 1.].
      unmatched_cls_target: a float32 tensor with shape [d_1, d_2, ..., d_k]
        which is consistent with the classification target for each
        anchor (and can be empty for scalar targets).  This shape must thus be
        compatible with the groundtruth labels that are passed to the "assign"
        function (which have shape [num_gt_boxes, d_1, d_2, ..., d_k]).
        If set to None, unmatched_cls_target is set to be [0] for each anchor.

    Raises:
      ValueError: if similarity_calc is not a RegionSimilarityCalculator or
        if matcher is not a Matcher or if box_coder is not a BoxCoder
    """
    self._similarity_calc = similarity_calc
    self._matcher = matcher
    self._box_coder = box_coder
    self._negative_class_weight = negative_class_weight
    if unmatched_cls_target is None:
      self._unmatched_cls_target = tf.constant([0], tf.float32)
    else:
      self._unmatched_cls_target = unmatched_cls_target

  @property
  def box_coder(self):
    return self._box_coder

  def assign(self, anchors, groundtruth_boxes, groundtruth_labels=None,
             groundtruth_weights=None, **params):
    """Assign classification and regression targets to each anchor.

    For a given set of anchors and groundtruth detections, match anchors
    to groundtruth_boxes and assign classification and regression targets to
    each anchor as well as weights based on the resulting match (specifying,
    e.g., which anchors should not contribute to training loss).

    Anchors that are not matched to anything are given a classification target
    of self._unmatched_cls_target which can be specified via the constructor.

    Args:
      anchors: a BoxList representing N anchors
      groundtruth_boxes: a BoxList representing M groundtruth boxes
      groundtruth_labels:  a tensor of shape [M, d_1, ... d_k]
        with labels for each of the ground_truth boxes. The subshape
        [d_1, ... d_k] can be empty (corresponding to scalar inputs).  When set
        to None, groundtruth_labels assumes a binary problem where all
        ground_truth boxes get a positive label (of 1).
      groundtruth_weights: a float tensor of shape [M] indicating the weight to
        assign to all anchors match to a particular groundtruth box. The weights
        must be in [0., 1.]. If None, all weights are set to 1.
      **params: Additional keyword arguments for specific implementations of
              the Matcher.

    Returns:
      cls_targets: a float32 tensor with shape [num_anchors, d_1, d_2 ... d_k],
        where the subshape [d_1, ..., d_k] is compatible with groundtruth_labels
        which has shape [num_gt_boxes, d_1, d_2, ... d_k].
      cls_weights: a float32 tensor with shape [num_anchors]
      reg_targets: a float32 tensor with shape [num_anchors, box_code_dimension]
      reg_weights: a float32 tensor with shape [num_anchors]
      match: a matcher.Match object encoding the match between anchors and
        groundtruth boxes, with rows corresponding to groundtruth boxes
        and columns corresponding to anchors.

    Raises:
      ValueError: if anchors or groundtruth_boxes are not of type
        box_list.BoxList
    """
    if not isinstance(anchors, box_list.BoxList):
      raise ValueError('anchors must be an BoxList')
    if not isinstance(groundtruth_boxes, box_list.BoxList):
      raise ValueError('groundtruth_boxes must be an BoxList')

    if groundtruth_labels is None:
      groundtruth_labels = tf.ones(tf.expand_dims(groundtruth_boxes.num_boxes(),
                                                  0))
      groundtruth_labels = tf.expand_dims(groundtruth_labels, -1)
    unmatched_shape_assert = shape_utils.assert_shape_equal(
        shape_utils.combined_static_and_dynamic_shape(groundtruth_labels)[1:],
        shape_utils.combined_static_and_dynamic_shape(
            self._unmatched_cls_target))
    labels_and_box_shapes_assert = shape_utils.assert_shape_equal(
        shape_utils.combined_static_and_dynamic_shape(
            groundtruth_labels)[:1],
        shape_utils.combined_static_and_dynamic_shape(
            groundtruth_boxes.get())[:1])

    if groundtruth_weights is None:
      num_gt_boxes = groundtruth_boxes.num_boxes_static()
      if not num_gt_boxes:
        num_gt_boxes = groundtruth_boxes.num_boxes()
      groundtruth_weights = tf.ones([num_gt_boxes], dtype=tf.float32)
    with tf.control_dependencies(
        [unmatched_shape_assert, labels_and_box_shapes_assert]):
      match_quality_matrix = self._similarity_calc.compare(groundtruth_boxes,
                                                           anchors)
      match = self._matcher.match(match_quality_matrix, **params)
      reg_targets = self._create_regression_targets(anchors,
                                                    groundtruth_boxes,
                                                    match)
      cls_targets = self._create_classification_targets(groundtruth_labels,
                                                        match)
      reg_weights = self._create_regression_weights(match, groundtruth_weights)
      cls_weights = self._create_classification_weights(match,
                                                        groundtruth_weights)

    num_anchors = anchors.num_boxes_static()
    if num_anchors is not None:
      reg_targets = self._reset_target_shape(reg_targets, num_anchors)
      cls_targets = self._reset_target_shape(cls_targets, num_anchors)
      reg_weights = self._reset_target_shape(reg_weights, num_anchors)
      cls_weights = self._reset_target_shape(cls_weights, num_anchors)

    return cls_targets, cls_weights, reg_targets, reg_weights, match

  def _reset_target_shape(self, target, num_anchors):
    """Sets the static shape of the target.

    Args:
      target: the target tensor. Its first dimension will be overwritten.
      num_anchors: the number of anchors, which is used to override the target's
        first dimension.

    Returns:
      A tensor with the shape info filled in.
    """
    target_shape = target.get_shape().as_list()
    target_shape[0] = num_anchors
    target.set_shape(target_shape)
    return target

  def _create_regression_targets(self, anchors, groundtruth_boxes, match):
    """Returns a regression target for each anchor.

    Args:
      anchors: a BoxList representing N anchors
      groundtruth_boxes: a BoxList representing M groundtruth_boxes
      match: a matcher.Match object

    Returns:
      reg_targets: a float32 tensor with shape [N, box_code_dimension]
    """
    matched_gt_boxes = match.gather_based_on_match(
        groundtruth_boxes.get(),
        unmatched_value=tf.zeros(4),
        ignored_value=tf.zeros(4))
    matched_gt_boxlist = box_list.BoxList(matched_gt_boxes)
    if groundtruth_boxes.has_field(KEYPOINTS_FIELD_NAME):
      groundtruth_keypoints = groundtruth_boxes.get_field(KEYPOINTS_FIELD_NAME)
      matched_keypoints = match.gather_based_on_match(
          groundtruth_keypoints,
          unmatched_value=tf.zeros(groundtruth_keypoints.get_shape()[1:]),
          ignored_value=tf.zeros(groundtruth_keypoints.get_shape()[1:]))
      matched_gt_boxlist.add_field(KEYPOINTS_FIELD_NAME, matched_keypoints)
    matched_reg_targets = self._box_coder.encode(matched_gt_boxlist, anchors)
    match_results_shape = shape_utils.combined_static_and_dynamic_shape(
        match.match_results)

    # Zero out the unmatched and ignored regression targets.
    unmatched_ignored_reg_targets = tf.tile(
        self._default_regression_target(), [match_results_shape[0], 1])
    matched_anchors_mask = match.matched_column_indicator()
    reg_targets = tf.where(matched_anchors_mask,
                           matched_reg_targets,
                           unmatched_ignored_reg_targets)
    return reg_targets

  def _default_regression_target(self):
    """Returns the default target for anchors to regress to.

    Default regression targets are set to zero (though in
    this implementation what these targets are set to should
    not matter as the regression weight of any box set to
    regress to the default target is zero).

    Returns:
      default_target: a float32 tensor with shape [1, box_code_dimension]
    """
    return tf.constant([self._box_coder.code_size*[0]], tf.float32)

  def _create_classification_targets(self, groundtruth_labels, match):
    """Create classification targets for each anchor.

    Assign a classification target of for each anchor to the matching
    groundtruth label that is provided by match.  Anchors that are not matched
    to anything are given the target self._unmatched_cls_target

    Args:
      groundtruth_labels:  a tensor of shape [num_gt_boxes, d_1, ... d_k]
        with labels for each of the ground_truth boxes. The subshape
        [d_1, ... d_k] can be empty (corresponding to scalar labels).
      match: a matcher.Match object that provides a matching between anchors
        and groundtruth boxes.

    Returns:
      a float32 tensor with shape [num_anchors, d_1, d_2 ... d_k], where the
      subshape [d_1, ..., d_k] is compatible with groundtruth_labels which has
      shape [num_gt_boxes, d_1, d_2, ... d_k].
    """
    return match.gather_based_on_match(
        groundtruth_labels,
        unmatched_value=self._unmatched_cls_target,
        ignored_value=self._unmatched_cls_target)

  def _create_regression_weights(self, match, groundtruth_weights):
    """Set regression weight for each anchor.

    Only positive anchors are set to contribute to the regression loss, so this
    method returns a weight of 1 for every positive anchor and 0 for every
    negative anchor.

    Args:
      match: a matcher.Match object that provides a matching between anchors
        and groundtruth boxes.
      groundtruth_weights: a float tensor of shape [M] indicating the weight to
        assign to all anchors match to a particular groundtruth box.

    Returns:
      a float32 tensor with shape [num_anchors] representing regression weights.
    """
    return match.gather_based_on_match(
        groundtruth_weights, ignored_value=0., unmatched_value=0.)

  def _create_classification_weights(self,
                                     match,
                                     groundtruth_weights):
    """Create classification weights for each anchor.

    Positive (matched) anchors are associated with a weight of
    positive_class_weight and negative (unmatched) anchors are associated with
    a weight of negative_class_weight. When anchors are ignored, weights are set
    to zero. By default, both positive/negative weights are set to 1.0,
    but they can be adjusted to handle class imbalance (which is almost always
    the case in object detection).

    Args:
      match: a matcher.Match object that provides a matching between anchors
        and groundtruth boxes.
      groundtruth_weights: a float tensor of shape [M] indicating the weight to
        assign to all anchors match to a particular groundtruth box.

    Returns:
      a float32 tensor with shape [num_anchors] representing classification
      weights.
    """
    return match.gather_based_on_match(
        groundtruth_weights,
        ignored_value=0.,
        unmatched_value=self._negative_class_weight)

  def get_box_coder(self):
    """Get BoxCoder of this TargetAssigner.

    Returns:
      BoxCoder object.
    """
    return self._box_coder



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Read weights stored in TensorFlow.js-format binary files."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import io
import os

import numpy as np
from tensorflowjs import quantization

_INPUT_DTYPES = [np.float16, np.float32, np.int32, np.complex64,
                 np.uint8, np.uint16, np.object, np.bool]

# Number of bytes used to encode the length of a string in a string tensor.
STRING_LENGTH_NUM_BYTES = 4
# The data type used to encode the length of a string in a string tensor.
STRING_LENGTH_DTYPE = np.dtype('uint32').newbyteorder('<')

def read_weights(weights_manifest, base_path, flatten=False):
  """Load weight values according to a TensorFlow.js weights manifest.

  Args:
    weights_manifest: A TensorFlow.js-format weights manifest (a JSON array).
    base_path: Base path prefix for the weights files.
    flatten: Whether all the weight groups in the return value are to be
      flattened as a single weights group. Default: `False`.

  Returns:
    If `flatten` is `False`, a `list` of weight groups. Each group is an array
    of weight entries. Each entry is a dict that maps a unique name to a numpy
    array, for example:
        entry = {
          'name': 'weight1',
          'data': np.array([1, 2, 3], 'float32')
        }

        Weights groups would then look like:
        weight_groups = [
          [group_0_entry1, group_0_entry2],
          [group_1_entry1, group_1_entry2],
        ]
    If `flatten` is `True`, returns a single weight group.
  """
  if not isinstance(weights_manifest, list):
    raise ValueError(
        'weights_manifest should be a `list`, but received %s' %
        type(weights_manifest))

  data_buffers = []
  for group in weights_manifest:
    buff = io.BytesIO()
    buff_writer = io.BufferedWriter(buff)
    for path in group['paths']:
      with open(os.path.join(base_path, path), 'rb') as f:
        buff_writer.write(f.read())
    buff_writer.flush()
    buff_writer.seek(0)
    data_buffers.append(buff.read())
  return decode_weights(weights_manifest, data_buffers, flatten=flatten)


def _deserialize_string_array(data_buffer, offset, shape):
  """Deserializes bytes into np.array of dtype `object` which holds strings.

  Each string value is preceded by 4 bytes which denote a 32-bit unsigned
  integer in little endian that specifies the byte length of the following
  string. This is followed by the actual string bytes. If the tensor has no
  strings there will be no bytes reserved. Empty strings will still take 4 bytes
  for the length.

  For example, a tensor that has 2 strings will be encoded as
  [byte length of s1][bytes of s1...][byte length of s2][bytes of s2...]

  where byte length always takes 4 bytes.

  Args:
    data_buffer: A buffer of bytes containing the serialized data.
    offset: The byte offset in that buffer that denotes the start of the tensor.
    shape: The logical shape of the tensor.

  Returns:
    A tuple of (np.array, offset) where the np.array contains the encoded
    strings, and the offset contains the new offset (the byte position in the
    buffer at the end of the string data).
  """
  size = int(np.prod(shape))
  if size == 0:
    return (np.array([], 'object').reshape(shape),
            offset + STRING_LENGTH_NUM_BYTES)
  vals = []
  for _ in range(size):
    byte_length = np.frombuffer(
        data_buffer[offset:offset + STRING_LENGTH_NUM_BYTES],
        STRING_LENGTH_DTYPE)[0]
    offset += STRING_LENGTH_NUM_BYTES
    string = data_buffer[offset:offset + byte_length]
    vals.append(string)
    offset += byte_length
  return np.array(vals, 'object').reshape(shape), offset


def _deserialize_numeric_array(data_buffer, offset, dtype, shape):
  weight_numel = 1
  for dim in shape:
    weight_numel *= dim
  return np.frombuffer(
      data_buffer, dtype=dtype, count=weight_numel,
      offset=offset).reshape(shape)

def decode_weights(weights_manifest, data_buffers, flatten=False):
  """Load weight values from buffer(s) according to a weights manifest.

  Args:
    weights_manifest: A TensorFlow.js-format weights manifest (a JSON array).
    data_buffers: A buffer or a `list` of buffers containing the weights values
      in binary format, concatenated in the order specified in
      `weights_manifest`. If a `list` of buffers, the length of the `list`
      must match the length of `weights_manifest`. A single buffer is
      interpreted as a `list` of one buffer and is valid only if the length of
      `weights_manifest` is `1`.
    flatten: Whether all the weight groups in the return value are to be
      flattened as a single weight groups. Default: `False`.

  Returns:
    If `flatten` is `False`, a `list` of weight groups. Each group is an array
    of weight entries. Each entry is a dict that maps a unique name to a numpy
    array, for example:
        entry = {
          'name': 'weight1',
          'data': np.array([1, 2, 3], 'float32')
        }

        Weights groups would then look like:
        weight_groups = [
          [group_0_entry1, group_0_entry2],
          [group_1_entry1, group_1_entry2],
        ]
    If `flatten` is `True`, returns a single weight group.

  Raises:
    ValueError: if the lengths of `weights_manifest` and `data_buffers` do not
      match.
  """
  if not isinstance(data_buffers, list):
    data_buffers = [data_buffers]
  if len(weights_manifest) != len(data_buffers):
    raise ValueError(
        'Mismatch in the length of weights_manifest (%d) and the length of '
        'data buffers (%d)' % (len(weights_manifest), len(data_buffers)))

  out = []
  for group, data_buffer in zip(weights_manifest, data_buffers):
    offset = 0
    out_group = []

    for weight in group['weights']:
      quant_info = weight.get('quantization', None)
      name = weight['name']
      if weight['dtype'] == 'string':
        # String array.
        dtype = np.object
      elif quant_info:
        # Quantized array.
        dtype = np.dtype(quant_info['dtype'])
      else:
        # Regular numeric array.
        dtype = np.dtype(weight['dtype'])
      shape = weight['shape']
      if dtype not in _INPUT_DTYPES:
        raise NotImplementedError('Unsupported data type: %s' % dtype)
      if weight['dtype'] == 'string':
        value, offset = _deserialize_string_array(data_buffer, offset, shape)
      else:
        value = _deserialize_numeric_array(data_buffer, offset, dtype, shape)
        offset += dtype.itemsize * value.size
      if quant_info:
        value = quantization.dequantize_weights(
            value, quant_info, np.dtype(weight['dtype']))
      out_group.append({'name': name, 'data': value})

    if flatten:
      out += out_group
    else:
      out.append(out_group)

  return out



# Lint as: python3
#   Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Customized metrics."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


def _get_batch_similarities(batch_label, full_vocab_similarities):
  return tf.gather(
      full_vocab_similarities, tf.transpose(batch_label)[0], axis=1)


class BatchRecall(tf.keras.metrics.Recall):
  """Compute batch recall for top_k."""

  def __init__(self, top_k=1, name='batch_recall'):
    super().__init__(name=name, top_k=top_k)

  def update_state(self, y_true, y_pred, sample_weight=None):
    """Update state of the metric.

    Args:
      y_true: the true labels with shape [batch_size, 1].
      y_pred: model output, which is the similarity matrix with shape
        [batch_size, label_embedding_vocab_size] between context and full vocab
        label embeddings.
      sample_weight: Optional weighting of each example. Defaults to 1.
    """
    label_indices = tf.eye(
        tf.shape(y_pred)[0], tf.shape(y_pred)[0], dtype=tf.dtypes.float32)
    # Since tf.keras default recall metric only accept predicted values in
    # range [0, 1], so use tf.nn.softmax to preprocess.
    normalized_logits = tf.nn.softmax(_get_batch_similarities(y_true, y_pred))
    super().update_state(label_indices, normalized_logits, sample_weight)


class GlobalRecall(tf.keras.metrics.Recall):
  """Compute global recall for top_k."""

  def __init__(self, top_k=1, name='global_recall'):
    super().__init__(name=name, top_k=top_k)

  def update_state(self, y_true, y_pred, sample_weight=None):
    """Update state of the metric.

    Args:
      y_true: the true labels with shape [batch_size, 1].
      y_pred: model output, which is the similarity matrix with shape
        [batch_size, label_embedding_vocab_size] between context and full vocab
        label embeddings.
      sample_weight: Optional weighting of each example. Defaults to 1.
    """
    label_indices = tf.one_hot(
        tf.transpose(y_true)[0], tf.shape(y_pred)[1], dtype=tf.dtypes.float32)
    # Since tf.keras default recall metric only accept predicted values in
    # range [0, 1], so use tf.nn.softmax to preprocess.
    normalized_logits = tf.nn.softmax(y_pred)
    super().update_state(label_indices, normalized_logits, sample_weight)


class BatchMeanRank(tf.keras.metrics.Mean):
  """Keras metric computing mean rank of correct label within batch."""

  def __init__(self, name='batch_mean_rank', **kwargs):
    super().__init__(name=name, **kwargs)

  def update_state(self, y_true, y_pred, sample_weight=None):
    """Update state of the metric.

    Args:
      y_true: the true labels with shape [batch_size, 1].
      y_pred: model output, which is the similarity matrix with shape
        [batch_size, label_embedding_vocab_size] between context and full vocab
        label embeddings. Hence to compute batch mean rank, the default global
        similarities will be coverted to batch similarities first with shape
        [batch_size, batch_size].
      sample_weight: Optional weighting of each example. Defaults to 1.
    """
    similarities = _get_batch_similarities(y_true, y_pred)
    # Get the ranks of the correct label for each context.
    batch_size = tf.shape(similarities)[0]
    label_indices = tf.expand_dims(tf.range(batch_size), -1)

    # Get the indices of similar labels in sorted order for each context.
    sorted_similarities = tf.argsort(
        similarities, axis=-1, direction='DESCENDING')

    ranks = tf.where(tf.equal(sorted_similarities, label_indices))[:, -1]
    ranks = tf.keras.backend.cast(ranks, 'float32')
    super().update_state(ranks, sample_weight=sample_weight)


class GlobalMeanRank(tf.keras.metrics.Mean):
  """Keras metric computing mean rank of correct label globally."""

  def __init__(self, name='global_mean_rank', **kwargs):
    super().__init__(name=name, **kwargs)

  def update_state(self, y_true, y_pred, sample_weight=None):
    """Update state of the metric.

    Args:
      y_true: the true labels with shape [batch_size, 1].
      y_pred: model output, which is the similarity matrix with shape
        [batch_size, label_embedding_vocab_size] between context and full vocab
        label embeddings.
      sample_weight: Optional weighting of each example. Defaults to 1.
    """
    similarities = y_pred
    label_indices = tf.expand_dims(tf.cast(y_true, dtype=tf.int32), -1)

    # Get the indices of similar labels in sorted order for each context.
    sorted_similarities = tf.argsort(
        similarities, axis=-1, direction='DESCENDING')

    ranks = tf.where(tf.equal(sorted_similarities, label_indices))[:, -1]
    ranks = tf.keras.backend.cast(ranks, 'float32')
    super().update_state(ranks, sample_weight=sample_weight)



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""APIs to train an object detection model."""

import os
import tempfile
from typing import Dict, Optional, Tuple, TypeVar

import tensorflow as tf
from tensorflow_examples.lite.model_maker.core import compat
from tensorflow_examples.lite.model_maker.core.api.api_util import mm_export
from tensorflow_examples.lite.model_maker.core.data_util import object_detector_dataloader
from tensorflow_examples.lite.model_maker.core.export_format import ExportFormat
from tensorflow_examples.lite.model_maker.core.task import configs
from tensorflow_examples.lite.model_maker.core.task import custom_model
from tensorflow_examples.lite.model_maker.core.task import model_spec as ms
from tensorflow_examples.lite.model_maker.core.task.metadata_writers.object_detector import metadata_writer_for_object_detector as metadata_writer
from tensorflow_examples.lite.model_maker.core.task.model_spec import object_detector_spec

from tensorflow_examples.lite.model_maker.third_party.efficientdet.keras import label_util

T = TypeVar('T', bound='ObjectDetector')


def _get_model_info(
    model_spec: object_detector_spec.EfficientDetModelSpec,
    quantization_config: Optional[configs.QuantizationConfig] = None,
) -> metadata_writer.ModelSpecificInfo:
  """Gets the specific info for the object detection model."""

  # Gets image_min/image_max for float/quantized model.
  image_min = -1
  image_max = 1
  if quantization_config:
    if quantization_config.inference_input_type == tf.uint8:
      image_min = 0
      image_max = 255
    elif quantization_config.inference_input_type == tf.int8:
      image_min = -128
      image_max = 127

  def _get_list(v):
    if isinstance(v, list) or isinstance(v, tuple):
      return v
    else:
      return [v]

  return metadata_writer.ModelSpecificInfo(
      name=model_spec.model_name,
      version='v1',
      image_width=model_spec.config.image_size[1],
      image_height=model_spec.config.image_size[0],
      image_min=image_min,
      image_max=image_max,
      mean=_get_list(model_spec.config.mean_rgb),
      std=_get_list(model_spec.config.stddev_rgb))


@mm_export('object_detector.ObjectDetector')
class ObjectDetector(custom_model.CustomModel):
  """ObjectDetector class for inference and exporting to tflite."""

  ALLOWED_EXPORT_FORMAT = (ExportFormat.TFLITE, ExportFormat.SAVED_MODEL,
                           ExportFormat.LABEL)

  def __init__(
      self,
      model_spec: object_detector_spec.EfficientDetModelSpec,
      label_map: Dict[int, str],
      representative_data: Optional[
          object_detector_dataloader.DataLoader] = None
  ) -> None:
    """Initializes the ObjectDetector class.

    Args:
      model_spec: Specification for the model.
      label_map:  Dict, map label integer ids to string label names such as {1:
        'person', 2: 'notperson'}. 0 is the reserved key for `background` and
          doesn't need to be included in `label_map`. Label names can't be
          duplicated.
      representative_data:  Representative dataset for full integer
        quantization. Used when converting the keras model to the TFLite model
        with full interger quantization.
    """
    super().__init__(model_spec, shuffle=None)
    if model_spec.config.label_map and model_spec.config.label_map != label_map:
      tf.compat.v1.logging.warn(
          'Label map is not the same as the previous label_map in model_spec.')
    model_spec.config.label_map = label_map
    # TODO(yuqili): num_classes = 1 have some issues during training. Thus we
    # make minimum num_classes=2 for now.
    model_spec.config.num_classes = max(2, max(label_map.keys()))
    self.representative_data = representative_data

  def create_model(self) -> tf.keras.Model:
    self.model = self.model_spec.create_model()
    return self.model

  def _get_dataset_and_steps(
      self,
      data: object_detector_dataloader.DataLoader,
      batch_size: int,
      is_training: bool,
  ) -> Tuple[Optional[tf.data.Dataset], int, Optional[str]]:
    """Gets dataset, steps and annotations json file."""
    if not data:
      return None, 0, None
    # TODO(b/171449557): Put this into DataLoader.
    dataset = data.gen_dataset(
        self.model_spec, batch_size, is_training=is_training)
    steps = len(data) // batch_size
    return dataset, steps, data.annotations_json_file

  def train(self,
            train_data: object_detector_dataloader.DataLoader,
            validation_data: Optional[
                object_detector_dataloader.DataLoader] = None,
            epochs: Optional[int] = None,
            batch_size: Optional[int] = None) -> tf.keras.Model:
    """Feeds the training data for training."""
    if not self.model_spec.config.drop_remainder:
      raise ValueError('Must set `drop_remainder=True` during training. '
                       'Otherwise it will fail.')

    batch_size = batch_size if batch_size else self.model_spec.batch_size
    # TODO(b/171449557): Upstream this to the parent class.
    if len(train_data) < batch_size:
      raise ValueError('The size of the train_data (%d) couldn\'t be smaller '
                       'than batch_size (%d). To solve this problem, set '
                       'the batch_size smaller or increase the size of the '
                       'train_data.' % (len(train_data), batch_size))
    if validation_data and len(validation_data) < batch_size:
      tf.compat.v1.logging.warn(
          'The size of the validation_data (%d) is smaller than batch_size '
          '(%d). Ignore the validation_data.' %
          (len(validation_data), batch_size))
      validation_data = None

    with self.model_spec.ds_strategy.scope():
      self.create_model()
      train_ds, steps_per_epoch, _ = self._get_dataset_and_steps(
          train_data, batch_size, is_training=True)
      validation_ds, validation_steps, val_json_file = self._get_dataset_and_steps(
          validation_data, batch_size, is_training=False)
      return self.model_spec.train(self.model, train_ds, steps_per_epoch,
                                   validation_ds, validation_steps, epochs,
                                   batch_size, val_json_file)

  def evaluate(self,
               data: object_detector_dataloader.DataLoader,
               batch_size: Optional[int] = None) -> Dict[str, float]:
    """Evaluates the model."""
    batch_size = batch_size if batch_size else self.model_spec.batch_size
    # Not to drop the smaller batch to evaluate the whole dataset.
    self.model_spec.config.drop_remainder = False
    ds = data.gen_dataset(self.model_spec, batch_size, is_training=False)
    steps = (len(data) + batch_size - 1) // batch_size
    # TODO(b/171449557): Upstream this to the parent class.
    if steps <= 0:
      raise ValueError('The size of the validation_data (%d) couldn\'t be '
                       'smaller than batch_size (%d). To solve this problem, '
                       'set the batch_size smaller or increase the size of the '
                       'validation_data.' % (len(data), batch_size))

    eval_metrics = self.model_spec.evaluate(self.model, ds, steps,
                                            data.annotations_json_file)
    # Set back drop_remainder=True since it must be True during training.
    # Otherwise it will fail.
    self.model_spec.config.drop_remainder = True
    return eval_metrics

  def evaluate_tflite(
      self, tflite_filepath: str,
      data: object_detector_dataloader.DataLoader) -> Dict[str, float]:
    """Evaluate the TFLite model."""
    ds = data.gen_dataset(self.model_spec, batch_size=1, is_training=False)
    return self.model_spec.evaluate_tflite(tflite_filepath, ds, len(data),
                                           data.annotations_json_file)

  def _export_saved_model(self, saved_model_dir: str) -> None:
    """Saves the model to Tensorflow SavedModel."""
    self.model_spec.export_saved_model(self.model, saved_model_dir)

  def _export_tflite(
      self,
      tflite_filepath: str,
      quantization_config: configs.QuantizationConfigType = 'default',
      with_metadata: bool = True,
      export_metadata_json_file: bool = False) -> None:
    """Converts the retrained model to tflite format and saves it.

    Args:
      tflite_filepath: File path to save tflite model.
      quantization_config: Configuration for post-training quantization. If
        'default', sets the `quantization_config` by default according to
        `self.model_spec`. If None, exports the float tflite model without
        quantization.
      with_metadata: Whether the output tflite model contains metadata.
      export_metadata_json_file: Whether to export metadata in json file. If
        True, export the metadata in the same directory as tflite model.Used
        only if `with_metadata` is True.
    """
    if quantization_config == 'default':
      quantization_config = self.model_spec.get_default_quantization_config(
          self.representative_data)

    self.model_spec.export_tflite(self.model, tflite_filepath,
                                  quantization_config)

    if with_metadata:
      with tempfile.TemporaryDirectory() as temp_dir:
        tf.compat.v1.logging.info(
            'Label file is inside the TFLite model with metadata.')
        label_filepath = os.path.join(temp_dir, 'labelmap.txt')
        self._export_labels(label_filepath)
        model_info = _get_model_info(self.model_spec, quantization_config)
        export_dir = os.path.dirname(tflite_filepath)
        populator = metadata_writer.MetadataPopulatorForObjectDetector(
            tflite_filepath, export_dir, model_info, label_filepath)
        populator.populate(export_metadata_json_file)

  def _export_labels(self, label_filepath: str) -> None:
    """Export labels to label_filepath."""
    tf.compat.v1.logging.info('Saving labels in %s.', label_filepath)
    num_classes = self.model_spec.config.num_classes
    label_map = label_util.get_label_map(self.model_spec.config.label_map)
    with tf.io.gfile.GFile(label_filepath, 'w') as f:
      # Ignores label_map[0] that's the background. The labels in the label file
      # for TFLite metadata should start from the actual labels without the
      # background.
      for i in range(num_classes):
        label = label_map[i + 1] if i + 1 in label_map else '???'
        f.write(label + '\n')

  @classmethod
  def create(cls,
             train_data: object_detector_dataloader.DataLoader,
             model_spec: object_detector_spec.EfficientDetModelSpec,
             validation_data: Optional[
                 object_detector_dataloader.DataLoader] = None,
             epochs: Optional[object_detector_dataloader.DataLoader] = None,
             batch_size: Optional[int] = None,
             train_whole_model: bool = False,
             do_train: bool = True) -> T:
    """Loads data and train the model for object detection.

    Args:
      train_data: Training data.
      model_spec: Specification for the model.
      validation_data: Validation data. If None, skips validation process.
      epochs: Number of epochs for training.
      batch_size: Batch size for training.
      train_whole_model: Boolean, False by default. If true, train the whole
        model. Otherwise, only train the layers that are not match
        `model_spec.config.var_freeze_expr`.
      do_train: Whether to run training.

    Returns:
      An instance based on ObjectDetector.
    """
    model_spec = ms.get(model_spec)
    if epochs is not None:
      model_spec.config.num_epochs = epochs
    if batch_size is not None:
      model_spec.config.batch_size = batch_size
    if train_whole_model:
      model_spec.config.var_freeze_expr = None
    if compat.get_tf_behavior() not in model_spec.compat_tf_versions:
      raise ValueError('Incompatible versions. Expect {}, but got {}.'.format(
          model_spec.compat_tf_versions, compat.get_tf_behavior()))

    object_detector = cls(model_spec, train_data.label_map, train_data)

    if do_train:
      tf.compat.v1.logging.info('Retraining the models...')
      object_detector.train(train_data, validation_data, epochs, batch_size)
    else:
      object_detector.create_model()

    return object_detector


# Shortcut function.
create = ObjectDetector.create
mm_export('object_detector.create').export_constant(__name__, 'create')



#   Copyright 2021 The TensorFlow Authors. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Prepare TF.Examples for on-device recommendation model.

Following functions are included: 1) downloading raw data 2) processing to user
activity sequence and splitting to train/test data 3) convert to TF.Examples
and write in output location.

More information about the movielens dataset can be found here:
https://grouplens.org/datasets/movielens/
"""

import collections
import json
import os
import random
import re

from absl import app
from absl import flags
from absl import logging
import pandas as pd
import tensorflow as tf

FLAGS = flags.FLAGS

# Permalinks to download movielens data.
MOVIELENS_1M_URL = "http://files.grouplens.org/datasets/movielens/ml-1m.zip"
MOVIELENS_ZIP_FILENAME = "ml-1m.zip"
MOVIELENS_ZIP_HASH = "a6898adb50b9ca05aa231689da44c217cb524e7ebd39d264c56e2832f2c54e20"
MOVIELENS_EXTRACTED_DIR = "ml-1m"
RATINGS_FILE_NAME = "ratings.dat"
MOVIES_FILE_NAME = "movies.dat"
RATINGS_DATA_COLUMNS = ["UserID", "MovieID", "Rating", "Timestamp"]
MOVIES_DATA_COLUMNS = ["MovieID", "Title", "Genres"]
OUTPUT_TRAINING_DATA_FILENAME = "train_movielens_1m.tfrecord"
OUTPUT_TESTING_DATA_FILENAME = "test_movielens_1m.tfrecord"
OUTPUT_MOVIE_VOCAB_FILENAME = "movie_vocab.json"
OUTPUT_MOVIE_YEAR_VOCAB_FILENAME = "movie_year_vocab.txt"
OUTPUT_MOVIE_GENRE_VOCAB_FILENAME = "movie_genre_vocab.txt"
OUTPUT_MOVIE_TITLE_UNIGRAM_VOCAB_FILENAME = "movie_title_unigram_vocab.txt"
OUTPUT_MOVIE_TITLE_BIGRAM_VOCAB_FILENAME = "movie_title_bigram_vocab.txt"
PAD_MOVIE_ID = 0
PAD_RATING = 0.0
PAD_MOVIE_YEAR = 0
UNKNOWN_STR = "UNK"
VOCAB_MOVIE_ID_INDEX = 0
VOCAB_COUNT_INDEX = 3


def define_flags():
  """Define flags."""
  flags.DEFINE_string("data_dir", "/tmp",
                      "Path to download and store movielens data.")
  flags.DEFINE_string("output_dir", None,
                      "Path to the directory of output files.")
  flags.DEFINE_bool("build_vocabs", True,
                    "If yes, generate movie feature vocabs.")
  flags.DEFINE_integer("min_timeline_length", 3,
                       "The minimum timeline length to construct examples.")
  flags.DEFINE_integer("max_context_length", 10,
                       "The maximum length of user context history.")
  flags.DEFINE_integer("max_context_movie_genre_length", 10,
                       "The maximum length of user context history.")
  flags.DEFINE_integer(
      "min_rating", None, "Minimum rating of movie that will be used to in "
      "training data")
  flags.DEFINE_float("train_data_fraction", 0.9, "Fraction of training data.")


class MovieInfo(
    collections.namedtuple(
        "MovieInfo", ["movie_id", "timestamp", "rating", "title", "genres"])):
  """Data holder of basic information of a movie."""
  __slots__ = ()

  def __new__(cls,
              movie_id=PAD_MOVIE_ID,
              timestamp=0,
              rating=PAD_RATING,
              title="",
              genres=""):
    return super(MovieInfo, cls).__new__(cls, movie_id, timestamp, rating,
                                         title, genres)


def download_and_extract_data(data_directory,
                              url=MOVIELENS_1M_URL,
                              fname=MOVIELENS_ZIP_FILENAME,
                              file_hash=MOVIELENS_ZIP_HASH,
                              extracted_dir_name=MOVIELENS_EXTRACTED_DIR):
  """Download and extract zip containing MovieLens data to a given directory.

  Args:
    data_directory: Local path to extract dataset to.
    url: Direct path to MovieLens dataset .zip file. See constants above for
      examples.
    fname: str, zip file name to download.
    file_hash: str, SHA-256 file hash.
    extracted_dir_name: str, extracted dir name under data_directory.

  Returns:
    Downloaded and extracted data file directory.
  """
  if not tf.io.gfile.exists(data_directory):
    tf.io.gfile.makedirs(data_directory)
  path_to_zip = tf.keras.utils.get_file(
      fname=fname,
      origin=url,
      file_hash=file_hash,
      hash_algorithm="sha256",
      extract=True,
      cache_dir=data_directory)
  extracted_file_dir = os.path.join(
      os.path.dirname(path_to_zip), extracted_dir_name)
  return extracted_file_dir


def read_data(data_directory, min_rating=None):
  """Read movielens ratings.dat and movies.dat file into dataframe."""
  ratings_df = pd.read_csv(
      os.path.join(data_directory, RATINGS_FILE_NAME),
      sep="::",
      names=RATINGS_DATA_COLUMNS,
      encoding="unicode_escape")  # May contain unicode. Need to escape.
  ratings_df["Timestamp"] = ratings_df["Timestamp"].apply(int)
  if min_rating is not None:
    ratings_df = ratings_df[ratings_df["Rating"] >= min_rating]
  movies_df = pd.read_csv(
      os.path.join(data_directory, MOVIES_FILE_NAME),
      sep="::",
      names=MOVIES_DATA_COLUMNS,
      encoding="unicode_escape")  # May contain unicode. Need to escape.
  return ratings_df, movies_df


def convert_to_timelines(ratings_df):
  """Convert ratings data to user."""
  timelines = collections.defaultdict(list)
  movie_counts = collections.Counter()
  for user_id, movie_id, rating, timestamp in ratings_df.values:
    timelines[user_id].append(
        MovieInfo(movie_id=movie_id, timestamp=int(timestamp), rating=rating))
    movie_counts[movie_id] += 1
  # Sort per-user timeline by timestamp
  for (user_id, context) in timelines.items():
    context.sort(key=lambda x: x.timestamp)
    timelines[user_id] = context
  return timelines, movie_counts


def generate_movies_dict(movies_df):
  """Generates movies dictionary from movies dataframe."""
  movies_dict = {
      movie_id: MovieInfo(movie_id=movie_id, title=title, genres=genres)
      for movie_id, title, genres in movies_df.values
  }
  movies_dict[0] = MovieInfo()
  return movies_dict


def extract_year_from_title(title):
  year = re.search(r"\((\d{4})\)", title)
  if year:
    return int(year.group(1))
  return 0


def generate_feature_of_movie_years(movies_dict, movies):
  """Extracts year feature for movies from movie title."""
  return [
      extract_year_from_title(movies_dict[movie.movie_id].title)
      for movie in movies
  ]


def generate_movie_genres(movies_dict, movies):
  """Create a feature of the genre of each movie.

  Save genre as a feature for the movies.

  Args:
    movies_dict: Dict of movies, keyed by movie_id with value of (title, genre)
    movies: list of movies to extract genres.

  Returns:
    movie_genres: list of genres of all input movies.
  """
  movie_genres = []
  for movie in movies:
    if not movies_dict[movie.movie_id].genres:
      continue
    genres = [
        tf.compat.as_bytes(genre)
        for genre in movies_dict[movie.movie_id].genres.split("|")
    ]
    movie_genres.extend(genres)

  return movie_genres


def _pad_or_truncate_movie_feature(feature, max_len, pad_value):
  feature.extend([pad_value for _ in range(max_len - len(feature))])
  return feature[:max_len]


def generate_examples_from_single_timeline(timeline,
                                           movies_dict,
                                           max_context_len=100,
                                           max_context_movie_genre_len=320):
  """Generate TF examples from a single user timeline.

  Generate TF examples from a single user timeline. Timeline with length less
  than minimum timeline length will be skipped. And if context user history
  length is shorter than max_context_len, features will be padded with default
  values.

  Args:
    timeline: The timeline to generate TF examples from.
    movies_dict: Dictionary of all MovieInfos.
    max_context_len: The maximum length of the context. If the context history
      length is less than max_context_length, features will be padded with
      default values.
    max_context_movie_genre_len: The length of movie genre feature.

  Returns:
    examples: Generated examples from this single timeline.
  """
  examples = []
  for label_idx in range(1, len(timeline)):
    start_idx = max(0, label_idx - max_context_len)
    context = timeline[start_idx:label_idx]
    # Pad context with out-of-vocab movie id 0.
    while len(context) < max_context_len:
      context.append(MovieInfo())
    label_movie_id = int(timeline[label_idx].movie_id)
    context_movie_id = [int(movie.movie_id) for movie in context]
    context_movie_rating = [movie.rating for movie in context]
    context_movie_year = generate_feature_of_movie_years(movies_dict, context)
    context_movie_genres = generate_movie_genres(movies_dict, context)
    context_movie_genres = _pad_or_truncate_movie_feature(
        context_movie_genres, max_context_movie_genre_len,
        tf.compat.as_bytes(UNKNOWN_STR))
    feature = {
        "context_movie_id":
            tf.train.Feature(
                int64_list=tf.train.Int64List(value=context_movie_id)),
        "context_movie_rating":
            tf.train.Feature(
                float_list=tf.train.FloatList(value=context_movie_rating)),
        "context_movie_genre":
            tf.train.Feature(
                bytes_list=tf.train.BytesList(value=context_movie_genres)),
        "context_movie_year":
            tf.train.Feature(
                int64_list=tf.train.Int64List(value=context_movie_year)),
        "label_movie_id":
            tf.train.Feature(
                int64_list=tf.train.Int64List(value=[label_movie_id]))
    }
    tf_example = tf.train.Example(features=tf.train.Features(feature=feature))
    examples.append(tf_example)

  return examples


def generate_examples_from_timelines(timelines,
                                     movies_df,
                                     min_timeline_len=3,
                                     max_context_len=100,
                                     max_context_movie_genre_len=320,
                                     train_data_fraction=0.9,
                                     random_seed=None,
                                     shuffle=True):
  """Convert user timelines to tf examples.

  Convert user timelines to tf examples by adding all possible context-label
  pairs in the examples pool.

  Args:
    timelines: The user timelines to process.
    movies_df: The dataframe of all movies.
    min_timeline_len: The minimum length of timeline. If the timeline length is
      less than min_timeline_len, empty examples list will be returned.
    max_context_len: The maximum length of the context. If the context history
      length is less than max_context_length, features will be padded with
      default values.
    max_context_movie_genre_len: The length of movie genre feature.
    train_data_fraction: Fraction of training data.
    random_seed: Seed for randomization.
    shuffle: Whether to shuffle the examples before splitting train and test
      data.

  Returns:
    train_examples: TF example list for training.
    test_examples: TF example list for testing.
  """
  examples = []
  movies_dict = generate_movies_dict(movies_df)
  progress_bar = tf.keras.utils.Progbar(len(timelines))
  for timeline in timelines.values():
    if len(timeline) < min_timeline_len:
      progress_bar.add(1)
      continue
    single_timeline_examples = generate_examples_from_single_timeline(
        timeline=timeline,
        movies_dict=movies_dict,
        max_context_len=max_context_len,
        max_context_movie_genre_len=max_context_movie_genre_len)
    examples.extend(single_timeline_examples)
    progress_bar.add(1)
  # Split the examples into train, test sets.
  if shuffle:
    random.seed(random_seed)
    random.shuffle(examples)
  last_train_index = round(len(examples) * train_data_fraction)

  train_examples = examples[:last_train_index]
  test_examples = examples[last_train_index:]
  return train_examples, test_examples


def generate_movie_feature_vocabs(movies_df, movie_counts):
  """Generate vocabularies for movie features.

  Generate vocabularies for movie features (movie_id, genre, year), sorted by
  usage count. Vocab id 0 will be reserved for default padding value.

  Args:
    movies_df: Dataframe for movies.
    movie_counts: Counts that each movie is rated.

  Returns:
    movie_id_vocab: List of all movie ids paired with movie usage count, and
      sorted by counts.
    movie_genre_vocab: List of all movie genres, sorted by genre usage counts.
    movie_year_vocab: List of all movie years, sorted by year usage counts.
  """
  movie_vocab = []
  movie_genre_counter = collections.Counter()
  movie_year_counter = collections.Counter()
  for movie_id, title, genres in movies_df.values:
    count = movie_counts.get(movie_id) or 0
    movie_vocab.append([movie_id, title, genres, count])
    year = extract_year_from_title(title)
    movie_year_counter[year] += 1
    for genre in genres.split("|"):
      movie_genre_counter[genre] += 1

  movie_vocab.sort(key=lambda x: x[VOCAB_COUNT_INDEX], reverse=True)  # by count
  movie_year_vocab = [0] + [x for x, _ in movie_year_counter.most_common()]
  movie_genre_vocab = [UNKNOWN_STR
                      ] + [x for x, _ in movie_genre_counter.most_common()]

  return (movie_vocab, movie_year_vocab, movie_genre_vocab)


def write_tfrecords(tf_examples, filename):
  """Writes tf examples to tfrecord file, and returns the count."""
  with tf.io.TFRecordWriter(filename) as file_writer:
    length = len(tf_examples)
    progress_bar = tf.keras.utils.Progbar(length)
    for example in tf_examples:
      file_writer.write(example.SerializeToString())
      progress_bar.add(1)
    return length


def write_vocab_json(vocab, filename):
  """Write generated movie vocabulary to specified file."""
  with open(filename, "w", encoding="utf-8") as jsonfile:
    json.dump(vocab, jsonfile, indent=2)


def write_vocab_txt(vocab, filename):
  with open(filename, "w", encoding="utf-8") as f:
    for item in vocab:
      f.write(str(item) + "\n")


def generate_datasets(extracted_data_dir,
                      output_dir,
                      min_timeline_length,
                      max_context_length,
                      max_context_movie_genre_length,
                      min_rating=None,
                      build_vocabs=True,
                      train_data_fraction=0.9,
                      train_filename=OUTPUT_TRAINING_DATA_FILENAME,
                      test_filename=OUTPUT_TESTING_DATA_FILENAME,
                      vocab_filename=OUTPUT_MOVIE_VOCAB_FILENAME,
                      vocab_year_filename=OUTPUT_MOVIE_YEAR_VOCAB_FILENAME,
                      vocab_genre_filename=OUTPUT_MOVIE_GENRE_VOCAB_FILENAME):
  """Generates train and test datasets as TFRecord, and returns stats."""
  logging.info("Reading data to dataframes.")
  ratings_df, movies_df = read_data(extracted_data_dir, min_rating=min_rating)
  logging.info("Generating movie rating user timelines.")
  timelines, movie_counts = convert_to_timelines(ratings_df)
  logging.info("Generating train and test examples.")
  train_examples, test_examples = generate_examples_from_timelines(
      timelines=timelines,
      movies_df=movies_df,
      min_timeline_len=min_timeline_length,
      max_context_len=max_context_length,
      max_context_movie_genre_len=max_context_movie_genre_length,
      train_data_fraction=train_data_fraction)

  if not tf.io.gfile.exists(output_dir):
    tf.io.gfile.makedirs(output_dir)
  logging.info("Writing generated training examples.")
  train_file = os.path.join(output_dir, train_filename)
  train_size = write_tfrecords(tf_examples=train_examples, filename=train_file)
  logging.info("Writing generated testing examples.")
  test_file = os.path.join(output_dir, test_filename)
  test_size = write_tfrecords(tf_examples=test_examples, filename=test_file)
  stats = {
      "train_size": train_size,
      "test_size": test_size,
      "train_file": train_file,
      "test_file": test_file,
  }

  if build_vocabs:
    (movie_vocab, movie_year_vocab, movie_genre_vocab) = (
        generate_movie_feature_vocabs(
            movies_df=movies_df, movie_counts=movie_counts))
    vocab_file = os.path.join(output_dir, vocab_filename)
    write_vocab_json(movie_vocab, filename=vocab_file)
    stats.update({
        "vocab_size": len(movie_vocab),
        "vocab_file": vocab_file,
        "vocab_max_id": max([arr[VOCAB_MOVIE_ID_INDEX] for arr in movie_vocab])
    })

    for vocab, filename, key in zip([movie_year_vocab, movie_genre_vocab],
                                    [vocab_year_filename, vocab_genre_filename],
                                    ["year_vocab", "genre_vocab"]):
      vocab_file = os.path.join(output_dir, filename)
      write_vocab_txt(vocab, filename=vocab_file)
      stats.update({
          key + "_size": len(vocab),
          key + "_file": vocab_file,
      })

  return stats


def main(_):
  logging.info("Downloading and extracting data.")
  extracted_data_dir = download_and_extract_data(data_directory=FLAGS.data_dir)

  stats = generate_datasets(
      extracted_data_dir=extracted_data_dir,
      output_dir=FLAGS.output_dir,
      min_timeline_length=FLAGS.min_timeline_length,
      max_context_length=FLAGS.max_context_length,
      max_context_movie_genre_length=FLAGS.max_context_movie_genre_length,
      min_rating=FLAGS.min_rating,
      build_vocabs=FLAGS.build_vocabs,
      train_data_fraction=FLAGS.train_data_fraction,
  )
  logging.info("Generated dataset: %s", stats)


if __name__ == "__main__":
  define_flags()
  app.run(main)
